@@ -20,6 +20,8 @@ pub type MultimapValue = Vec<u8>;
 pub type MultimapKeyValue = (MultimapKey, MultimapValue);
 pub type MultimapKeyValues = BTreeSet<(EntryHash, MultimapKeyValue)>;
 
+const MULTIMAP_REMOVED_MARK: &[u8] = b"";
+
 impl Safe {
     /// Create a Multimap on the network
     pub async fn multimap_create(
@@ -27,13 +29,18 @@ impl Safe {
         name: Option<XorName>,
         type_tag: u64,
         private: bool,
+        dry_run: bool,
     ) -> Result<XorUrl> {
         debug!("Creating a Multimap");
-        let xorname = self
+        let (xorname, op_batch) = self
             .safe_client
             .create_register(name, type_tag, None, private)
             .await?;
 
+        if !dry_run {
+            self.safe_client.apply_register_ops(op_batch).await?;
+        }
+
         let scope = if private {
             Scope::Private
         } else {
@@ -90,6 +97,7 @@ impl Safe {
         multimap_url: &str,
         entry: MultimapKeyValue,
         replace: BTreeSet<EntryHash>,
+        dry_run: bool,
     ) -> Result<EntryHash> {
         debug!("Inserting '{:?}' into Multimap at {}", entry, multimap_url);
         let serialised_entry = rmp_serde::to_vec_named(&entry).map_err(|err| {
@@ -111,14 +119,56 @@ impl Safe {
                 )))
             }
         };
-        self.safe_client
+
+        let (entry_hash, op_batch) = self
+            .safe_client
             .write_to_register(address, data, replace)
-            .await
+            .await?;
+
+        if !dry_run {
+            self.safe_client.apply_register_ops(op_batch).await?;
+        }
+
+        Ok(entry_hash)
+    }
+
+    /// Remove entries from a Multimap on the network
+    /// This tombstones the removed entries, effectively hiding them if they where the latest
+    /// Note that they are still stored on the network as history is kept,
+    /// and you can still access them with their EntryHash
+    pub async fn multimap_remove(
+        &self,
+        url: &str,
+        to_remove: BTreeSet<EntryHash>,
+        dry_run: bool,
+    ) -> Result<EntryHash> {
+        debug!("Removing from Multimap at {}: {:?}", url, to_remove);
+        let safeurl = Safe::parse_url(url)?;
+        let address = match safeurl.address() {
+            DataAddress::Register(reg_address) => reg_address,
+            other => {
+                return Err(Error::InvalidXorUrl(format!(
+                    "The multimap url {} has an {:?} address.\
+                    To remove an entry from a multimap, the address must be a register address.",
+                    url, other
+                )))
+            }
+        };
+        let (entry_hash, op_batch) = self
+            .safe_client
+            .write_to_register(address, MULTIMAP_REMOVED_MARK.to_vec(), to_remove)
+            .await?;
+
+        if !dry_run {
+            self.safe_client.apply_register_ops(op_batch).await?;
+        }
+
+        Ok(entry_hash)
     }
 
     // Crate's helper to return the value of a Multimap on
     // the network without resolving the Url,
-    // optionally filtering by hash and/or key.
+    // filtering by hash if a version is provided
     pub(crate) async fn fetch_multimap_values(&self, safeurl: &Url) -> Result<MultimapKeyValues> {
         let entries = match self.register_fetch_entries(safeurl).await {
             Ok(data) => {
@@ -127,17 +177,22 @@ impl Safe {
             }
             Err(Error::EmptyContent(_)) => Err(Error::EmptyContent(format!(
                 "Multimap found at \"{}\" was empty",
-                safeurl
+                safeurl.to_string()
+            ))),
+            Err(Error::ContentNotFound(_)) => Err(Error::ContentNotFound(format!(
+                "No Multimap found at \"{}\"",
+                safeurl.to_string()
             ))),
-            Err(Error::ContentNotFound(_)) => Err(Error::ContentNotFound(
-                "No Multimap found at this address".to_string(),
-            )),
             other => other,
         }?;
 
         // We parse each entry in the Register as a 'MultimapKeyValue'
         let mut multimap_key_vals = MultimapKeyValues::new();
         for (hash, entry) in entries.iter() {
+            if entry == MULTIMAP_REMOVED_MARK {
+                // this is a tombstone entry created to delete some old entries
+                continue;
+            }
             let key_val = Self::decode_multimap_entry(entry)?;
             multimap_key_vals.insert((*hash, key_val));
         }
@@ -167,8 +222,16 @@ impl Safe {
             Err(other) => Err(other),
         }?;
 
-        let key_val = Self::decode_multimap_entry(&entry)?;
-        Ok(key_val)
+        // We parse the entry in the Register as a 'MultimapKeyValue'
+        if entry == MULTIMAP_REMOVED_MARK {
+            Err(Error::EmptyContent(format!(
+                "Entry found at \"{}\" is a tombstone (deletion marker)",
+                safeurl
+            )))
+        } else {
+            let key_val = Self::decode_multimap_entry(&entry)?;
+            Ok(key_val)
+        }
     }
 
     fn decode_multimap_entry(entry: &[u8]) -> Result<MultimapKeyValue> {
@@ -179,7 +242,7 @@ impl Safe {
 
 #[cfg(test)]
 mod tests {
-    use crate::{app::test_helpers::new_safe_instance, retry_loop_for_pattern};
+    use crate::{app::test_helpers::new_safe_instance, retry_loop, retry_loop_for_pattern};
     use anyhow::Result;
     use std::collections::BTreeSet;
 
@@ -187,8 +250,8 @@ mod tests {
     async fn test_multimap_create() -> Result<()> {
         let safe = new_safe_instance().await?;
 
-        let xorurl = safe.multimap_create(None, 25_000, false).await?;
-        let xorurl_priv = safe.multimap_create(None, 25_000, true).await?;
+        let xorurl = safe.multimap_create(None, 25_000, false, false).await?;
+        let xorurl_priv = safe.multimap_create(None, 25_000, true, false).await?;
 
         let key = b"".to_vec();
         let received_data = safe.multimap_get_by_key(&xorurl, &key).await?;
@@ -210,17 +273,17 @@ mod tests {
         let val2 = b"value2".to_vec();
         let key_val2 = (key.clone(), val2.clone());
 
-        let xorurl = safe.multimap_create(None, 25_000, false).await?;
-        let xorurl_priv = safe.multimap_create(None, 25_000, true).await?;
+        let xorurl = safe.multimap_create(None, 25_000, false, false).await?;
+        let xorurl_priv = safe.multimap_create(None, 25_000, true, false).await?;
 
         let _ = safe.multimap_get_by_key(&xorurl, &key).await?;
         let _ = safe.multimap_get_by_key(&xorurl_priv, &key).await?;
 
         let hash = safe
-            .multimap_insert(&xorurl, key_val.clone(), BTreeSet::new())
+            .multimap_insert(&xorurl, key_val.clone(), BTreeSet::new(), false)
             .await?;
         let hash_priv = safe
-            .multimap_insert(&xorurl_priv, key_val.clone(), BTreeSet::new())
+            .multimap_insert(&xorurl_priv, key_val.clone(), BTreeSet::new(), false)
             .await?;
 
         let received_data = retry_loop_for_pattern!(safe.multimap_get_by_key(&xorurl, &key), Ok(v) if !v.is_empty())?;
@@ -238,11 +301,16 @@ mod tests {
         // Let's now test an insert which replace the previous value for a key
         let hashes_to_replace = vec![hash].into_iter().collect();
         let hash2 = safe
-            .multimap_insert(&xorurl, key_val2.clone(), hashes_to_replace)
+            .multimap_insert(&xorurl, key_val2.clone(), hashes_to_replace, false)
             .await?;
         let hashes_priv_to_replace = vec![hash_priv].into_iter().collect();
         let hash_priv2 = safe
-            .multimap_insert(&xorurl_priv, key_val2.clone(), hashes_priv_to_replace)
+            .multimap_insert(
+                &xorurl_priv,
+                key_val2.clone(),
+                hashes_priv_to_replace,
+                false,
+            )
             .await?;
 
         let received_data = retry_loop_for_pattern!(safe.multimap_get_by_key(&xorurl, &key),
@@ -272,24 +340,24 @@ mod tests {
         let val2 = b"value2".to_vec();
         let key_val2 = (key2.clone(), val2.clone());
 
-        let xorurl = safe.multimap_create(None, 25_000, false).await?;
-        let xorurl_priv = safe.multimap_create(None, 25_000, true).await?;
+        let xorurl = safe.multimap_create(None, 25_000, false, false).await?;
+        let xorurl_priv = safe.multimap_create(None, 25_000, true, false).await?;
 
         let _ = safe.multimap_get_by_key(&xorurl, &key).await?;
         let _ = safe.multimap_get_by_key(&xorurl_priv, &key).await?;
 
         let hash = safe
-            .multimap_insert(&xorurl, key_val.clone(), BTreeSet::new())
+            .multimap_insert(&xorurl, key_val.clone(), BTreeSet::new(), false)
             .await?;
         let hash2 = safe
-            .multimap_insert(&xorurl, key_val2.clone(), BTreeSet::new())
+            .multimap_insert(&xorurl, key_val2.clone(), BTreeSet::new(), false)
             .await?;
 
         let hash_priv = safe
-            .multimap_insert(&xorurl_priv, key_val.clone(), BTreeSet::new())
+            .multimap_insert(&xorurl_priv, key_val.clone(), BTreeSet::new(), false)
             .await?;
         let hash_priv2 = safe
-            .multimap_insert(&xorurl_priv, key_val2.clone(), BTreeSet::new())
+            .multimap_insert(&xorurl_priv, key_val2.clone(), BTreeSet::new(), false)
             .await?;
 
         let received_data = safe.multimap_get_by_hash(&xorurl, hash).await?;
@@ -306,4 +374,66 @@ mod tests {
 
         Ok(())
     }
+
+    #[tokio::test]
+    async fn test_multimap_remove() -> Result<()> {
+        let safe = new_safe_instance().await?;
+        let key = b"key".to_vec();
+        let val = b"value".to_vec();
+        let key_val = (key.clone(), val.clone());
+
+        let xorurl = safe.multimap_create(None, 25_000, false, false).await?;
+        let xorurl_priv = safe.multimap_create(None, 25_000, true, false).await?;
+
+        let _ = retry_loop!(safe.multimap_get_by_key(&xorurl, &key));
+        let _ = retry_loop!(safe.multimap_get_by_key(&xorurl_priv, &key));
+
+        let hash = safe
+            .multimap_insert(&xorurl, key_val.clone(), BTreeSet::new(), false)
+            .await?;
+        let hash_priv = safe
+            .multimap_insert(&xorurl_priv, key_val, BTreeSet::new(), false)
+            .await?;
+
+        let received_data = retry_loop_for_pattern!(safe.multimap_get_by_key(&xorurl, &key), Ok(v) if !v.is_empty())?;
+        let received_data_priv = retry_loop_for_pattern!(safe.multimap_get_by_key(&xorurl_priv, &key), Ok(v) if !v.is_empty())?;
+
+        assert_eq!(received_data.len(), 1);
+        let (read_hash, read_key_val) = received_data.into_iter().next().unwrap();
+        assert_eq!(
+            (read_hash, read_key_val),
+            (hash, (key.clone(), val.clone()))
+        );
+
+        let hashes_to_remove = vec![hash].into_iter().collect();
+        let removed_mark_hash = safe
+            .multimap_remove(&xorurl, hashes_to_remove, false)
+            .await?;
+        assert_ne!(removed_mark_hash, hash);
+
+        assert_eq!(
+            retry_loop_for_pattern!(safe.multimap_get_by_key(&xorurl, &key),
+                                    Ok(entries) if entries.is_empty())?,
+            Default::default()
+        );
+
+        assert_eq!(received_data_priv.len(), 1);
+        let (read_hash, read_key_val) = received_data_priv.into_iter().next().unwrap();
+        assert_eq!((read_hash, read_key_val), (hash_priv, (key.clone(), val)));
+
+        let hashes_to_remove = vec![hash_priv].into_iter().collect();
+        let removed_mark_hash = safe
+            .multimap_remove(&xorurl_priv, hashes_to_remove, false)
+            .await?;
+        assert_ne!(removed_mark_hash, hash_priv);
+
+        assert_eq!(
+            retry_loop_for_pattern!(
+                safe.multimap_get_by_key(&xorurl_priv, &key),
+                Ok(entries) if entries.is_empty())?,
+            Default::default()
+        );
+
+        Ok(())
+    }
 }
