@@ -9,14 +9,15 @@
 
 mod nrs_map;
 
-pub use crate::url::{ContentType, VersionHash};
+pub use crate::url::{ContentType, DataType, VersionHash};
 pub use nrs_map::NrsMap;
 
+use nrs_map::parse_out_subnames;
+
 use crate::{app::Safe, register::EntryHash, Error, Result, Url};
 
-use bytes::{Buf, Bytes};
-use log::{debug, info, warn};
-use std::collections::BTreeSet;
+use log::{debug, info};
+use std::collections::{BTreeMap, BTreeSet};
 use std::str;
 
 // Type tag to use for the NrsMapContainer stored on Register
@@ -35,9 +36,8 @@ impl Safe {
     ///            Public Name
     /// ```
     /// Registers the given NRS top name on the network.
-    /// Returns the versioned NRS Url (containing a VersionHash)
-    /// `safe://{top_name}?v={version_hash}`
-    /// Note that this NRS Url is not linked yet to anything. You just registered the topname here.
+    /// Returns the NRS Url: `safe://{top_name}
+    /// Note that this NRS Url is not linked to anything yet. You just registered the topname here.
     /// You can now associate public_names (with that topname) to links using `nrs_associate` or `nrs_add`
     pub async fn nrs_create(&self, top_name: &str, dry_run: bool) -> Result<Url> {
         info!("Creating an NRS map for: {}", top_name);
@@ -50,40 +50,25 @@ impl Safe {
             return Err(Error::NrsNameAlreadyExists(top_name.to_owned()));
         }
 
-        if dry_run {
-            return Err(Error::NotImplementedError(
-                "No dry run for nrs_create. Version info cannot be determined. (Register operations need this functionality implemented first.)".to_string(),
-            ));
-        }
-
-        // Create and store NrsMap for top_name in a Public Blob
-        let nrs_map = NrsMap::default();
-        let nrs_map_xorurl = self.store_nrs_map(&nrs_map).await?;
-
-        // Create a new Register and reference the NRS map's blob in the Register
-        let reg_xorurl = self
-            .register_create(Some(nrs_xorname), NRS_MAP_TYPE_TAG, false)
-            .await?;
-        let reg_entry = nrs_map_xorurl.as_bytes().to_vec();
-        // Note that we can use the higher level register API here
-        // because reg_xorurl is not an NRS url
-        // (high level register API uses resolution that uses NRS!)
-        let entry_hash = self
-            .register_write(&reg_xorurl, reg_entry, BTreeSet::new())
+        // Create a new empty Multimap
+        let _ = self
+            .multimap_create(Some(nrs_xorname), NRS_MAP_TYPE_TAG, false, dry_run)
             .await?;
 
-        Ok(get_versioned_nrs_url(url_str, entry_hash)?)
+        let mut url = Url::from_url(&url_str)?;
+        url.set_content_type(ContentType::NrsMapContainer)?;
+        Ok(url)
     }
 
     /// # Associates a public name to a link
     /// The top name of the input public name needs to be registered first with `nrs_create`
+    ///
     /// ```no_run
     /// safe://<subName>.<topName>/path/to/whatever?var=value
     ///        |-----------------|
     ///            Public Name
     /// ```
-    /// Associates the given public_name to the link, stores it in the NrsMap registered for the
-    /// public name's top name on the network.
+    /// Associates the given public_name to the link.
     /// Errors out if the topname is not registered.
     /// Returns the versioned NRS Url (containing a VersionHash) now pointing to the provided link:
     /// `safe://{public_name}?v={version_hash}`
@@ -93,29 +78,27 @@ impl Safe {
             public_name, link
         );
 
-        // check public_name
+        // check input
         let url_str = validate_nrs_public_name(public_name)?;
+        validate_nrs_url(link)?;
 
-        // fetch and edit nrs_map
-        let (mut nrs_map, version) = self.fetch_nrs_map(public_name, None).await?;
-        nrs_map.associate(public_name, link)?;
-        debug!("The new NRS Map: {:?}", nrs_map);
+        // get current latest for subname
+        let safe_url = Safe::parse_url(public_name)?;
+        let subname = parse_out_subnames(public_name);
+        let current_versions = self
+            .fetch_multimap_value_by_key(&safe_url, subname.as_bytes())
+            .await?
+            .into_iter()
+            .map(|(hash, _)| hash)
+            .collect();
 
-        if dry_run {
-            return Err(Error::NotImplementedError(
-                "No dry run for nrs_add. Version info cannot be determined. (Register operations need this functionality implemented first.)".to_string(),
-            ));
-        }
-
-        // store updated nrs_map
-        let nrs_map_xorurl = self.store_nrs_map(&nrs_map).await?;
-        let old_values: BTreeSet<EntryHash> = [version.entry_hash()].iter().copied().collect();
-        let reg_entry = nrs_map_xorurl.as_bytes().to_vec();
-        let safe_url = Safe::parse_url(&url_str)?;
-        let address = self.get_register_address(&safe_url)?;
+        // update with new entry
+        let entry = (
+            subname.as_bytes().to_vec(),
+            link.to_string().as_bytes().to_vec(),
+        );
         let entry_hash = self
-            .safe_client
-            .write_to_register(address, reg_entry, old_values)
+            .multimap_insert(&url_str, entry, current_versions, dry_run)
             .await?;
 
         Ok(get_versioned_nrs_url(url_str, entry_hash)?)
@@ -175,25 +158,19 @@ impl Safe {
         // check public_name
         let url_str = validate_nrs_public_name(public_name)?;
 
-        // fetch and edit nrs_map
-        let (mut nrs_map, version) = self.fetch_nrs_map(public_name, None).await?;
-        nrs_map.remove(public_name)?;
+        // get current latest for subname
+        let safe_url = Safe::parse_url(public_name)?;
+        let subname = parse_out_subnames(public_name);
+        let current_versions = self
+            .fetch_multimap_value_by_key(&safe_url, subname.as_bytes())
+            .await?
+            .into_iter()
+            .map(|(hash, _)| hash)
+            .collect();
 
-        if dry_run {
-            return Err(Error::NotImplementedError(
-                "No dry run for nrs_remove. Version info cannot be determined. (Register operations need this functionality implemented first.)".to_string(),
-            ));
-        }
-
-        // store updated nrs_map
-        let nrs_map_xorurl = self.store_nrs_map(&nrs_map).await?;
-        let old_values: BTreeSet<EntryHash> = [version.entry_hash()].iter().copied().collect();
-        let reg_entry = nrs_map_xorurl.as_bytes().to_vec();
-        let safe_url = Safe::parse_url(&url_str)?;
-        let address = self.get_register_address(&safe_url)?;
+        // remove
         let entry_hash = self
-            .safe_client
-            .write_to_register(address, reg_entry, old_values)
+            .multimap_remove(&url_str, current_versions, dry_run)
             .await?;
 
         Ok(get_versioned_nrs_url(url_str, entry_hash)?)
@@ -208,7 +185,11 @@ impl Safe {
     ///            Public Name
     /// ```
     /// Finds the Url associated with the given public name on the network.
-    /// Returns the associated Url for the given public name for that version along with the NrsMap
+    /// If multiple entries are found for the same public name, there's a conflict.
+    /// If there are conflicts for subnames other than the one requested, get proceeds as usual,
+    /// but the NrsMap returned will ignore those conflicts.
+    /// Otherwise, it returns an error.
+    /// Returns the associated Url for the given public name for that version along with an NrsMap
     pub async fn nrs_get(
         &self,
         public_name: &str,
@@ -219,7 +200,20 @@ impl Safe {
             public_name, version
         );
 
-        let (nrs_map, _version) = self.fetch_nrs_map(public_name, version).await?;
+        // get nrs_map, ignoring conflicting entries if they are not the ones we're getting
+        let nrs_map = match self.nrs_get_subnames_map(public_name, version).await {
+            Ok(map) => Ok(map),
+            Err(Error::ConflictingNrsEntries(str, conflicting_entries, map)) => {
+                let subname = parse_out_subnames(public_name);
+                if conflicting_entries.iter().any(|(sub, _)| sub == &subname) {
+                    Err(Error::ConflictingNrsEntries(str, conflicting_entries, map))
+                } else {
+                    Ok(map)
+                }
+            }
+            Err(e) => Err(e),
+        }?;
+
         let url = nrs_map.get(public_name)?;
         Ok((url, nrs_map))
     }
@@ -230,82 +224,86 @@ impl Safe {
         public_name: &str,
         version: Option<VersionHash>,
     ) -> Result<NrsMap> {
-        let (nrs_map, _version) = self.fetch_nrs_map(public_name, version).await?;
-        Ok(nrs_map)
-    }
+        // fetch multimap entries
+        let safe_url = Safe::parse_url(public_name)?;
+        let res = self.fetch_multimap_values(&safe_url).await;
+        let multimap_keyvals = match res {
+            Ok(s) => Ok(s),
+            Err(Error::EmptyContent(_)) => Ok(BTreeSet::new()),
+            Err(Error::ContentNotFound(e)) => Err(Error::ContentNotFound(format!(
+                "No Nrs Map entry found at {}: {}",
+                safe_url.to_string(),
+                e
+            ))),
+            Err(e) => Err(Error::NetDataError(format!(
+                "Failed to get Nrs Map entries: {}",
+                e
+            ))),
+        }?;
 
-    // Private helper function to fetch the nrs_map from the network
-    // If no version is provided, fetches the latest
-    // Always returns the version of the fetched content along with the NrsMap
-    async fn fetch_nrs_map(
-        &self,
-        public_name: &str,
-        version: Option<VersionHash>,
-    ) -> Result<(NrsMap, VersionHash)> {
-        // assign version
-        let mut safe_url = Safe::parse_url(public_name)?;
-        safe_url.set_content_version(version);
-        let safe_url = safe_url;
+        // collect a raw map with serialized data, get specific version if needed
+        let raw_set = match version {
+            Some(v) => {
+                let hash = v.entry_hash();
+                if multimap_keyvals
+                    .iter()
+                    .any(|(h, _)| VersionHash::from(h) == v)
+                {
+                    // just return the key val set if we have the version we're looking for
+                    multimap_keyvals
+                        .into_iter()
+                        .map(|(_hash, key_val)| key_val)
+                        .collect()
+                } else {
+                    // manually fetch the missing versionned entry
+                    let mut key_vals: BTreeSet<_> = multimap_keyvals
+                        .into_iter()
+                        .map(|(_hash, key_val)| key_val)
+                        .collect();
+                    let key_val = self.fetch_multimap_value_by_hash(&safe_url, hash).await?;
+                    let fetched_key = &key_val.0;
 
-        // fetch entries and wrap errors
-        let entries = self
-            .register_fetch_entries(&safe_url)
-            .await
-            .map_err(|e| match e {
-                Error::ContentNotFound(_) => {
-                    Error::ContentNotFound("No NRS Map found at this address".to_string())
+                    // remove other entries with the same key
+                    key_vals.retain(|(k, _)| k != fetched_key);
+
+                    // insert the versioned entry
+                    key_vals.insert(key_val);
+                    key_vals
                 }
-                Error::VersionNotFound(msg) => Error::VersionNotFound(msg),
-                err => Error::NetDataError(format!("Failed to get current version: {}", err)),
-            })?;
-
-        // take the 1st entry (TODO Multiple entries)
-        if entries.len() > 1 {
-            return Err(Error::RegisterFork("Multiple NRS map entries not managed, this happends when 2 clients write concurrently to a NRS map".to_string()));
-        }
-        let first_entry = entries.iter().next();
-        let (version, nrs_map_url) = match first_entry {
-            Some((entry_hash, url)) => (entry_hash.into(), Url::from_xorurl(str::from_utf8(url)?)?),
-            None => {
-                warn!(
-                    "NRS map Register found at XOR name \"{:?}\" was empty",
-                    safe_url.xorname()
-                );
-                return Ok((NrsMap::default(), VersionHash::default()));
             }
+            None => multimap_keyvals
+                .into_iter()
+                .map(|(_hash, key_val)| key_val)
+                .collect(),
         };
 
-        // Using the NrsMap url we can now fetch the NrsMap and deserialise it
-        let serialised_nrs_map = self.fetch_public_data(&nrs_map_url, None).await?;
+        // deserialize
+        let clean_set: BTreeSet<(String, Url)> = raw_set
+            .into_iter()
+            .map(|(subname_bytes, url_bytes)| {
+                let subname = str::from_utf8(&subname_bytes)?;
+                let url = Url::from_url(str::from_utf8(&url_bytes)?)?;
+                Ok((subname.to_owned(), url))
+            })
+            .collect::<Result<BTreeSet<(String, Url)>>>()?;
 
-        debug!("Nrs map v{} retrieved: {:?} ", version, &serialised_nrs_map);
-        let nrs_map = serde_json::from_str(&String::from_utf8_lossy(serialised_nrs_map.chunk()))
-            .map_err(|err| {
-                Error::ContentError(format!(
-                    "Couldn't deserialise the NrsMap stored in the NrsContainer: {:?}",
-                    err
-                ))
-            })?;
+        // turn into map
+        let subnames_map: BTreeMap<String, Url> = clean_set.clone().into_iter().collect();
+        let nrs_map = NrsMap {
+            map: subnames_map.clone(),
+        };
 
-        Ok((nrs_map, version))
-    }
+        // check for conflicting entries (same subname, different url)
+        let set_len = clean_set.len();
+        let map_len = subnames_map.len();
+        if map_len != set_len {
+            let set_from_map: BTreeSet<(String, Url)> = subnames_map.into_iter().collect();
+            let conflicting_entries: Vec<(String, Url)> =
+                clean_set.difference(&set_from_map).cloned().collect();
+            return Err(Error::ConflictingNrsEntries("Found multiple entries for the same name, this happens when 2 clients write concurrently to the same NRS mapping. It can be fixed by simply associating a new link to the conflicting names.".to_string(), conflicting_entries, nrs_map));
+        }
 
-    // Private helper to serialise an NrsMap and store it in a Public Blob
-    async fn store_nrs_map(&self, nrs_map: &NrsMap) -> Result<String> {
-        // The NrsMapContainer is a Register where each NRS Map version is
-        // an entry containing the XOR-URL of the Blob that contains the serialised NrsMap.
-        let serialised_nrs_map = serde_json::to_string(nrs_map).map_err(|err| {
-            Error::Serialisation(format!(
-                "Couldn't serialise the NrsMap generated: {:?}",
-                err
-            ))
-        })?;
-
-        let nrs_map_xorurl = self
-            .store_public_bytes(Bytes::from(serialised_nrs_map), None, false)
-            .await?;
-
-        Ok(nrs_map_xorurl)
+        Ok(nrs_map)
     }
 }
 
@@ -323,7 +321,7 @@ fn validate_nrs_top_name(top_name: &str) -> Result<String> {
     let safe_url = Safe::parse_url(&sanitised_url)?;
     if safe_url.top_name() != top_name {
         return Err(Error::InvalidInput(
-            format!("The NRS top name \"{}\" is invalid because it contains url parts, please remove any path, version, subnames, etc...", top_name)
+            format!("The NRS top name \"{}\" is invalid because it contains url parts. Please remove any path, version or subnames.", top_name)
         ));
     }
     Ok(sanitised_url)
@@ -335,12 +333,37 @@ fn validate_nrs_public_name(public_name: &str) -> Result<String> {
     let safe_url = Safe::parse_url(&sanitised_url)?;
     if safe_url.public_name() != public_name {
         return Err(Error::InvalidInput(
-            format!("The NRS public name \"{}\" is invalid because it contains url parts, please remove any path, version, etc...", public_name)
+            format!("The NRS public name \"{}\" is invalid because it contains url parts. Please remove any path or version.", public_name)
         ));
     }
     Ok(sanitised_url)
 }
 
+// helper function to check a Url used for NRS
+// - checks if the url is valid
+// - checks if it has a version if its data is versionable
+fn validate_nrs_url(link: &Url) -> Result<()> {
+    if link.content_version().is_none() {
+        let content_type = link.content_type();
+        let data_type = link.data_type();
+        if content_type == ContentType::FilesContainer
+            || content_type == ContentType::NrsMapContainer
+        {
+            return Err(Error::UnversionedContentError(format!(
+                "The linked content ({}) is versionable, therefore NRS requires the link to specify a hash: {}",
+                content_type, link.to_string()
+            )));
+        } else if data_type == DataType::Register {
+            return Err(Error::UnversionedContentError(format!(
+                "The linked content ({}) is versionable, therefore NRS requires the link to specify a hash: {}",
+                data_type, link.to_string()
+            )));
+        }
+    }
+
+    Ok(())
+}
+
 // Makes sure there’s a (and only one) "safe://" in front of input name
 fn sanitised_url(name: &str) -> String {
     format!("safe://{}", name.strip_prefix("safe://").unwrap_or(name))
@@ -351,10 +374,10 @@ mod tests {
     use super::*;
     use crate::{
         app::test_helpers::{new_safe_instance, random_nrs_name},
-        retry_loop, retry_loop_for_pattern, Url,
+        retry_loop, retry_loop_for_pattern, Error, Url,
     };
     use anyhow::{anyhow, bail, Result};
-    use std::str::FromStr;
+    use std::{matches, str::FromStr};
 
     #[tokio::test]
     async fn test_nrs_create() -> Result<()> {
@@ -444,7 +467,7 @@ mod tests {
             Err(Error::InvalidInput(msg)) => assert_eq!(
                 msg,
                 format!(
-                    "The NRS public name \"{}\" is invalid because it contains url parts, please remove any path, version, etc...",
+                    "The NRS public name \"{}\" is invalid because it contains url parts. Please remove any path or version.",
                     versioned_sitename
                 )
             ),
@@ -459,7 +482,7 @@ mod tests {
                 assert_eq!(
                     msg,
                     format!(
-                        "The NRS public name \"{}\" is invalid because it contains url parts, please remove any path, version, etc...",
+                        "The NRS public name \"{}\" is invalid because it contains url parts. Please remove any path or version.",
                         versioned_sitename
                     )
                 );
@@ -574,7 +597,7 @@ mod tests {
 
         // test with invalid top name
         let invalid_top_name = "atffdgasd/d";
-        let expected_err = format!("The NRS top name \"{}\" is invalid because it contains url parts, please remove any path, version, subnames, etc...", invalid_top_name);
+        let expected_err = format!("The NRS top name \"{}\" is invalid because it contains url parts. Please remove any path, version or subnames.", invalid_top_name);
         match safe.nrs_create(invalid_top_name, true).await {
             Ok(_) => bail!("Unexpected NRS success when expected to fail with invalid top name"),
             Err(Error::InvalidInput(e)) => assert_eq!(e, expected_err),
@@ -582,4 +605,75 @@ mod tests {
         };
         Ok(())
     }
+
+    #[tokio::test]
+    async fn test_nrs_conflicting_names() -> Result<()> {
+        let site_name = random_nrs_name();
+        let mut safe = new_safe_instance().await?;
+
+        // let's create an empty files container so we have a valid to link
+        let (link, _, _) = safe
+            .files_container_create(None, None, true, true, false)
+            .await?;
+        let (version0, _) = retry_loop!(safe.files_container_get(&link));
+
+        // associate a first name
+        let mut valid_link = Url::from_url(&link)?;
+        valid_link.set_content_version(Some(version0));
+
+        let (nrs_url, did_create) = retry_loop!(safe.nrs_add(&site_name, &valid_link, false));
+        assert!(did_create);
+
+        let _ = retry_loop!(safe.fetch(&nrs_url.to_string(), None));
+
+        // associate a second name
+        let second_valid_link = Url::from_url(&link)?;
+        valid_link.set_content_version(Some(version0));
+        let site_name2 = format!("sub.{}", &site_name);
+
+        let (nrs_url2, did_create) =
+            retry_loop!(safe.nrs_add(&site_name2, &second_valid_link, false));
+        assert!(!did_create);
+
+        let _ = retry_loop!(safe.fetch(&nrs_url2.to_string(), None));
+
+        // manually add a conflicting name
+        let another_valid_url = nrs_url;
+        let url_str = validate_nrs_top_name(&site_name)?;
+        let entry = (
+            "".as_bytes().to_vec(),
+            another_valid_url.to_string().as_bytes().to_vec(),
+        );
+        let _ = safe
+            .multimap_insert(&url_str, entry, BTreeSet::new(), false)
+            .await?;
+
+        // get of other name should be ok
+        let _ = retry_loop_for_pattern!(safe.nrs_get(&site_name2, None), Ok((res_url, _)) if res_url == &second_valid_link)?;
+
+        // get of conflicting name should error out
+        let conflict_error =
+            retry_loop_for_pattern!(safe.nrs_get(&site_name, None), Err(_) if true);
+        assert!(matches!(
+            conflict_error,
+            Err(Error::ConflictingNrsEntries { .. })
+        ));
+
+        // check for the error content
+        if let Err(Error::ConflictingNrsEntries(_, dups, _)) = conflict_error {
+            let got_entries: Result<()> = dups.into_iter().try_for_each(|(subname, url)| {
+                assert_eq!(subname, "");
+                assert!(url == valid_link || url == another_valid_url);
+                Ok(())
+            });
+            assert!(got_entries.is_ok());
+        }
+
+        // resolve the error
+        let _ = retry_loop!(safe.nrs_associate(&site_name, &valid_link, false));
+
+        // get should work now
+        let _ = retry_loop_for_pattern!(safe.nrs_get(&site_name, None), Ok((res_url, _)) if res_url == &valid_link)?;
+        Ok(())
+    }
 }
