@@ -8,161 +8,30 @@
 // Software.
 
 mod handlers;
+mod safe_data;
 
-use super::{
-    files::{FileInfo, FilesMap},
-    multimap::MultimapKeyValues,
-    nrs::NrsMap,
-    register::{Entry, EntryHash},
-    Safe, XorName,
-};
+use super::{files::FileInfo, Safe};
 pub use super::{ContentType, DataType, Url, VersionHash, XorUrlBase};
 use crate::{Error, Result};
-use async_recursion::async_recursion;
-use bytes::Bytes;
 use log::{debug, info};
-use serde::{Deserialize, Serialize};
-use std::collections::BTreeSet;
+pub use safe_data::SafeData;
 
 pub type Range = Option<(Option<u64>, Option<u64>)>;
 
 // Maximum number of indirections allowed when resolving a safe:// URL following links
 const INDIRECTION_LIMIT: usize = 10;
 
-/// SafeData contains the data types fetchable using the Safe Network resolver
-#[allow(clippy::large_enum_variant)]
-// FilesContainer is significantly larger than the other variants
-#[derive(Debug, PartialEq, Deserialize, Serialize, Clone)]
-pub enum SafeData {
-    SafeKey {
-        xorurl: String,
-        xorname: XorName,
-        resolved_from: String,
-    },
-    FilesContainer {
-        xorurl: String,
-        xorname: XorName,
-        type_tag: u64,
-        version: VersionHash,
-        files_map: FilesMap,
-        data_type: DataType,
-        metadata: Option<FileInfo>,
-        resolves_into: Option<Url>,
-        resolved_from: String,
-    },
-    PublicBlob {
-        xorurl: String,
-        xorname: XorName,
-        data: Bytes,
-        media_type: Option<String>,
-        metadata: Option<FileInfo>,
-        resolved_from: String,
-    },
-    NrsMapContainer {
-        public_name: Option<String>,
-        xorurl: String,
-        xorname: XorName,
-        type_tag: u64,
-        version: VersionHash,
-        nrs_map: NrsMap,
-        data_type: DataType,
-        resolves_into: Option<Url>,
-        resolved_from: String,
-    },
-    Multimap {
-        xorurl: String,
-        xorname: XorName,
-        type_tag: u64,
-        data: MultimapKeyValues,
-        resolved_from: String,
-    },
-    PublicRegister {
-        xorurl: String,
-        xorname: XorName,
-        type_tag: u64,
-        data: BTreeSet<(EntryHash, Entry)>,
-        resolved_from: String,
-    },
-    PrivateRegister {
-        xorurl: String,
-        xorname: XorName,
-        type_tag: u64,
-        data: BTreeSet<(EntryHash, Entry)>,
-        resolved_from: String,
-    },
-    // TODO? Suggestion: Error type for easy error tracking in a resolution chain?
-}
-
-impl SafeData {
-    pub fn xorurl(&self) -> String {
-        use SafeData::*;
-        match self {
-            SafeKey { xorurl, .. }
-            | FilesContainer { xorurl, .. }
-            | PublicBlob { xorurl, .. }
-            | NrsMapContainer { xorurl, .. }
-            | Multimap { xorurl, .. }
-            | PublicRegister { xorurl, .. }
-            | PrivateRegister { xorurl, .. } => xorurl.clone(),
-        }
-    }
-
-    pub fn resolved_from(&self) -> String {
-        use SafeData::*;
-        match self {
-            SafeKey { resolved_from, .. }
-            | FilesContainer { resolved_from, .. }
-            | PublicBlob { resolved_from, .. }
-            | NrsMapContainer { resolved_from, .. }
-            | Multimap { resolved_from, .. }
-            | PublicRegister { resolved_from, .. }
-            | PrivateRegister { resolved_from, .. } => resolved_from.clone(),
-        }
-    }
-
-    pub fn resolves_into(&self) -> Option<Url> {
-        use SafeData::*;
-        match self {
-            SafeKey { .. }
-            | PublicBlob { .. }
-            | Multimap { .. }
-            | PublicRegister { .. }
-            | PrivateRegister { .. } => None,
-            FilesContainer { resolves_into, .. } | NrsMapContainer { resolves_into, .. } => {
-                resolves_into.clone()
-            }
-        }
-    }
-
-    pub fn metadata(&self) -> Option<FileInfo> {
-        use SafeData::*;
-        match self {
-            SafeKey { .. }
-            | Multimap { .. }
-            | PublicRegister { .. }
-            | PrivateRegister { .. }
-            | NrsMapContainer { .. } => None,
-            FilesContainer { metadata, .. } | PublicBlob { metadata, .. } => metadata.clone(),
-        }
-    }
-}
-
 impl Safe {
     /// Parses a string URL "safe://url" and returns a safe URL
-    /// Recursively resolves until it reaches the final URL
+    /// Resolves until it reaches the final URL
     pub async fn parse_and_resolve_url(&self, url: &str) -> Result<Url> {
         let safe_url = Safe::parse_url(url)?;
         let orig_path = safe_url.path_decoded()?;
 
         // Obtain the resolution chain without resolving the URL's path
         let mut resolution_chain = self
-            .recursive_resolve_url(
-                safe_url,
-                None,
-                false,
-                None,
-                false, // don't resolve the URL's path
-                INDIRECTION_LIMIT,
+            .fully_resolve_url(
+                safe_url, None, false, None, false, // don't resolve the URL's path
             )
             .await?;
 
@@ -214,7 +83,7 @@ impl Safe {
         info!("URL parsed successfully, fetching: {}", url);
 
         let mut resolution_chain = self
-            .recursive_resolve_url(safe_url, None, true, range, true, INDIRECTION_LIMIT)
+            .fully_resolve_url(safe_url, None, true, range, true)
             .await?;
 
         resolution_chain
@@ -225,7 +94,8 @@ impl Safe {
     /// # Inspect a safe:// URL and retrieve metadata information but the actual target content
     /// # As opposed to 'fetch' function, the actual target content won't be fetched, and only
     /// # the URL will be inspected resolving it as necessary to find the target location.
-    /// # This is helpful if you are interested in knowing about the target content rather than
+    /// # This is helpful if you are interested in knowing about the target content,
+    /// # and/or each of the Url resolution steps taken to the target content, rather than
     /// # trying to revieve the actual content.
     ///
     /// ## Examples
@@ -266,16 +136,16 @@ impl Safe {
     pub async fn inspect(&mut self, url: &str) -> Result<Vec<SafeData>> {
         let safe_url = Safe::parse_url(url)?;
         info!("URL parsed successfully, inspecting: {}", url);
-        self.recursive_resolve_url(safe_url, None, false, None, true, INDIRECTION_LIMIT)
+        self.fully_resolve_url(safe_url, None, false, None, true)
             .await
     }
 
-    // Retrieves all pieces of data that resulted from resolving the given URL.
-    // Keeping a copy of the intermediary steps when indirections occur
-    // Recursively resolves the given URL until
+    // Retrieves all pieces of data that resulted from resolving the given URL,
+    // keeping a copy of the intermediary resolution steps when indirections occur.
+    // Resolves the given URL until
     // - it reaches the final piece of data
     // - or reaches the indirection limit
-    // Returns a Vector with the data for all steps
+    // Returns a Vector with the data for all resolution steps
     //
     // NB: When resolving a Blob, metadata can be attached to it (attached_metadata)
     // Blobs don't have metadata on SAFE but the FileContainers linking to them have it
@@ -286,59 +156,42 @@ impl Safe {
     // NB: recursive (resolutions that resolve to themselves) aren't managed but since the
     // indirections are limited, it's probably not worth the overhead check.
     // Will need it if we allow infinite indirections though.
-    #[async_recursion]
-    async fn recursive_resolve_url(
+    async fn fully_resolve_url(
         &self,
         input_url: Url,
         attached_metadata: Option<FileInfo>,
         retrieve_data: bool,
         range: Range,
         resolve_path: bool,
-        indirections_limit: usize,
     ) -> Result<Vec<SafeData>> {
         debug!(
             "Fetching URL: {} with content of type: {:?}, data type: {:?}",
-            input_url.to_string(),
+            input_url,
             input_url.content_type(),
             input_url.data_type()
         );
 
-        // fetch safe_data from URL
-        let safe_data = self
-            .resolve_url(
-                input_url,
-                attached_metadata,
-                retrieve_data,
-                range,
-                resolve_path,
-            )
-            .await?;
-        let next_step = safe_data.resolves_into();
+        let mut indirections_limit = INDIRECTION_LIMIT;
+        let mut safe_data_vec = vec![];
+        let mut next_step = Some(input_url);
+        let mut metadata = attached_metadata;
+        while let Some(next_url) = next_step {
+            // fetch safe_data from URL
+            let safe_data = self
+                .resolve_url(next_url, metadata, retrieve_data, range, resolve_path)
+                .await?;
 
-        // stop cases
-        let next_url;
-        match next_step {
-            None => return Ok(vec![safe_data]),
-            Some(url) => next_url = url,
-        }
-        if indirections_limit == 0 {
-            return Err(Error::ContentError(format!("The maximum number of indirections ({}) was reached when trying to resolve the URL provided", INDIRECTION_LIMIT)));
+            next_step = safe_data.resolves_into();
+            metadata = safe_data.metadata();
+            safe_data_vec.push(safe_data);
+
+            if indirections_limit == 0 {
+                return Err(Error::ContentError(format!("The maximum number of indirections ({}) was reached when trying to resolve the URL provided", INDIRECTION_LIMIT)));
+            }
+
+            indirections_limit -= 1;
         }
 
-        // fetch next (and attach current metadata to it)
-        let next_safe_data = self
-            .recursive_resolve_url(
-                next_url,
-                safe_data.metadata(),
-                retrieve_data,
-                range,
-                resolve_path,
-                indirections_limit - 1,
-            )
-            .await?;
-
-        let mut safe_data_vec = vec![safe_data];
-        safe_data_vec.extend(next_safe_data);
         Ok(safe_data_vec)
     }
 
@@ -348,31 +201,33 @@ impl Safe {
         &self,
         input_url: Url,
         attached_metadata: Option<FileInfo>,
-        fetch_mode: bool,
+        retrieve_data: bool,
         range: Range,
         resolve_path: bool,
     ) -> Result<SafeData> {
         debug!(
-            "Resolving URL: {}, of content type: {:?}, and data type: {:?}",
+            "Resolving URL: {}, of content type: {:?}, and data type: {:?}, address {:?}",
             input_url.to_xorurl_string(),
             input_url.content_type(),
-            input_url.data_type()
+            input_url.data_type(),
+            input_url.address()
         );
+
         match input_url.content_type() {
             ContentType::FilesContainer => {
                 self.resolve_file_container(input_url, resolve_path).await
             }
             ContentType::NrsMapContainer => self.resolve_nrs_map_container(input_url).await,
-            ContentType::Multimap => self.resolve_multimap(input_url, fetch_mode).await,
+            ContentType::Multimap => self.resolve_multimap(input_url, retrieve_data).await,
             ContentType::Raw => {
-                self.resolve_raw(input_url, attached_metadata, fetch_mode, range)
+                self.resolve_raw(input_url, attached_metadata, retrieve_data, range)
                     .await
             }
             ContentType::MediaType(media_type_str) => {
                 self.resolve_mediatype(
                     input_url,
                     attached_metadata,
-                    fetch_mode,
+                    retrieve_data,
                     range,
                     media_type_str,
                 )
@@ -390,6 +245,7 @@ mod tests {
     use super::*;
     use crate::{app::files, app::test_helpers::new_safe_instance, retry_loop, Scope, Url};
     use anyhow::{anyhow, bail, Context, Result};
+    use bytes::Bytes;
     use rand::{distributions::Alphanumeric, thread_rng, Rng};
     use safe_network::types::DataAddress;
     use std::io::Read;
@@ -621,8 +477,8 @@ mod tests {
         let fetch_first_half = Some((None, Some(size as u64 / 2)));
         let content = retry_loop!(safe.fetch(&xorurl, fetch_first_half));
 
-        if let SafeData::PublicBlob { data, .. } = &content {
-            assert_eq!(data.clone(), saved_data.slice(0..size / 2).to_vec());
+        if let SafeData::PublicBlob { data, .. } = content {
+            assert_eq!(data, saved_data.slice(0..size / 2));
         } else {
             bail!("Content fetched is not a PublicBlob: {:?}", content);
         }
@@ -631,8 +487,8 @@ mod tests {
         let fetch_second_half = Some((Some(size as u64 / 2), Some(size as u64)));
         let content = safe.fetch(&xorurl, fetch_second_half).await?;
 
-        if let SafeData::PublicBlob { data, .. } = &content {
-            assert_eq!(data.clone(), saved_data[size / 2..size].to_vec());
+        if let SafeData::PublicBlob { data, .. } = content {
+            assert_eq!(data, saved_data[size / 2..]);
             Ok(())
         } else {
             Err(anyhow!(
@@ -680,8 +536,8 @@ mod tests {
         // fetch first half and match
         let fetch_first_half = Some((Some(0), Some(file_size as u64 / 2)));
         let content = safe.fetch(&nrs_url, fetch_first_half).await?;
-        if let SafeData::PublicBlob { data, .. } = &content {
-            assert_eq!(data.clone(), file_data[0..file_size / 2].to_vec());
+        if let SafeData::PublicBlob { data, .. } = content {
+            assert_eq!(data, file_data[0..file_size / 2]);
         } else {
             bail!("Content fetched is not a PublicBlob: {:?}", content);
         }
@@ -689,8 +545,8 @@ mod tests {
         // fetch second half and match
         let fetch_second_half = Some((Some(file_size as u64 / 2), Some(file_size as u64)));
         let content = safe.fetch(&nrs_url, fetch_second_half).await?;
-        if let SafeData::PublicBlob { data, .. } = &content {
-            assert_eq!(data.clone(), file_data[file_size / 2..file_size].to_vec());
+        if let SafeData::PublicBlob { data, .. } = content {
+            assert_eq!(data, file_data[file_size / 2..]);
         } else {
             bail!("Content fetched is not a PublicBlob: {:?}", content);
         }
